"""module that calculates the truckfactor for a local git repository"""
import os
import shlex
import subprocess
import csv
import collections
import re
import itertools
import numpy as np
from pathlib import Path

def truck_factor_repo(local_path_repo, cur_wd, start_date, end_date):
    """
    Calculate the truckfactor from a repository and write the truckfactor statistics to a csv file.

    Parameters
    ----------
    local_path_repo : str
        Local path to the directory of the git repository.
    cur_wd : str
        Current working directory
    start_date : str
        Start date from commit history in ISO format.
    end_date
        End date from commit history in ISO format.

    Returns
    -------
    tf - int
        Number of truckfactor developers.
    """

    list_files = list_files_linguist(local_path_repo, cur_wd)  # get the list of files

    dict_authors_file = authors_file_list(
        local_path_repo, cur_wd, list_files, start_date, end_date
    )  # get the name of each author and their authored files

    truck_factor = calculate_tf(
        local_path_repo, cur_wd, dict_authors_file
    )  # calculate the truck factor

    return truck_factor


def list_files_linguist(local_path_repo, cur_wd):
    """
    Get a list of source files from a repository that disregards third party code, documentation, images, ...

    Parameters
    ----------
    local_path_repo : str
        Local path to the directory of the git repository.
    cur_wd : str
        Current working directory.

    Returns
    -------
    files_linguist : list - [filename]
        List of source files from the git repository.
    """

    git_cmd = shlex.split("github-linguist --breakdown")

    pathlib_curwd = Path(cur_wd)
    pathlib_repo = pathlib_curwd.joinpath(local_path_repo)

    os.chdir(str(pathlib_repo))
    process = subprocess.run(git_cmd, capture_output=True, encoding="utf-8", shell=True
    )  # run the shell command git_cmd
    os.chdir(str(pathlib_curwd))  # change back to the current working directory

    # divide the output of the git command into list entries
    files_linguist = process.stdout.split("\n")

    # filter files_linguist: remove entries which do not represent a file name
    files_linguist = list(
        filter(lambda file: check_remove(file) == True, files_linguist)
    )

    return files_linguist


def check_remove(string):
    """
    Return true if input string has "." in it and not "%" in it
    """

    return bool("." in string and "%" not in string)


def authors_file_list(local_path_repo, cur_wd, list_files, start_date, end_date):
    """
    Loop through each file in a repository:
    - extract the commit history for each file
    - define the authors of each file
    - create a dictionary with the authors name and append their authored files

    Returns
    -------
    dict_files_authors : dict - {author: [filename]}
        Dictionary that stores the authors and their authored files.
    """

    dict_files_authors = {}

    for filename in list_files:
        commit_history = get_commit_history(
            local_path_repo, cur_wd, filename, start_date, end_date
        )
        if (not commit_history):  # stops the current iteration and moves to the next iteration if commit_history is empty
            continue

        list_doa = doa(commit_history)
        file_authors = get_authors_file(list_doa, filename)

        for i in range(1, len(file_authors)):
            if (file_authors[i] in dict_files_authors):  # if author is in dictionary append the filenames
                dict_files_authors[file_authors[i]].append(file_authors[0])
            else:
                dict_files_authors[file_authors[i]] = [file_authors[0]]  # if author not yet in dictionary, create a new entry with the author and filenames

    return dict_files_authors


def get_commit_history(local_path_repo, cur_wd, filename, start_date, end_date):
    """
    Invoke subprocess and run git command 'log shortstat' to extract the commit history of a file

    Parameters
    ----------
    local_path_repo : str
        Local path to the directory of the git repository.
    cur_wd : str
        Current working directory
    filename : str
        Name of the source file.
    start_date : str
        Start date from commit history in ISO format.
    end_date
        End date from commit history in ISO format.

    Returns
    -------
    commit_info_file : list of lists - [[hash, author, date, insertions, deletions, total number of changes]]
        Commit history - information about commits of a source file.
    """

    git_cmd = [
        "git", "log", "--after={" + start_date + "}", 
        "--before={" + end_date + "}", "--pretty=format:Commit%n%H%n%an%n%as%n%ae%n%s%n", 
        "--shortstat", filename
        ]

    pathlib_curwd = Path(cur_wd)
    pathlib_repo = pathlib_curwd.joinpath(local_path_repo)

    os.chdir(str(pathlib_repo))
    process = subprocess.run(git_cmd, capture_output=True, encoding="utf-8", check=True)
    os.chdir(str(pathlib_curwd))  # change back to the current working directory

    log_shortstat = process.stdout.split("\n")

    # checks if the commit history is empty (no commits for a file in the specified timeline)
    if check_empty_commit_history(log_shortstat):
        return False

    # append the ouptut of git log to a list, each commit stat can be separated by the entry "Commit"
    commit_info_file = []
    i_sep_cur = 0
    i_sep_prev = 0
    for i in range(1, len(log_shortstat)):  # loop through list
        if (log_shortstat[i] == "Commit"):  # if 'Commit' is found set the index for the previous and current separators
            i_sep_prev = i_sep_cur
            i_sep_cur = i
            commit_info_file.append(
                log_shortstat[i_sep_prev : i_sep_cur - 1]
            )  # append all items between string "commit"
    commit_info_file.append(
        log_shortstat[i_sep_cur:-1]
    )  # append items after last string "commit"

    commit_info_file = [
        list(filter(None, commit)) for commit in commit_info_file
    ]  # filter None values

    # extract the insertions and deletions
    for commit in commit_info_file:
        try:
            re_insert = re.search(r"(\d+|$)(?=\s*insert)", commit[6])
            insertions = re_insert.group(0) if re_insert else 0
            re_delet = re.search(r"(\d+|$)(?=\s*delet)", commit[6])
            deletions = re_delet.group(0) if re_delet else 0
            commit[6] = [insertions, deletions, int(insertions) + int(deletions)]
        except:
            # if no changes are made, the insertions and deletions are not displayed, therefore we add them if no changes can be detected
            commit.append([0, 0, 0])

    return commit_info_file


def check_empty_commit_history(commit_history):
    """
    Return true if the commit_history is empty
    """

    return bool(len(commit_history) == 1 and not commit_history[0])


def doa(commit_history):
    """
    Calculate the degree of authorship for a source file for every author from a commit history

    Parameters
    ----------
    commit_history : str
        Commit history of a file (list of lists)

    Returns
    -------
    normalized_list : list of lists - [[author, doa, normalized doa]]
        Normalized degree-of-authorship list.
    """

    first_author = first_author_file(commit_history)  # get the first author of a file
    first_authorship = 0

    doa_list = []

    # loop through each commit in a file
    for commit in commit_history:
        author = commit[2]
        if author == first_author:  # set first authorship
            first_authorship = 1
        else:
            first_authorship = 0

        deliveries = commit[6][2]  # number of deliveries, number of changes made by the current developer in the loop
        acceptances = sum(
            commit_ac[6][2] for commit_ac in commit_history if commit_ac != author
        )  # number of acceptances, number of changes mady by any developer other than the current one
        degree_of_authorship = float(
            3.293 + 1.098 * first_authorship + 0.164 * deliveries - 0.321 * np.log(1 + acceptances)
        )  # calculate the degree of authorship for the current author in the loop
        doa_list.append([author, degree_of_authorship])

    dict_unique_author = collections.defaultdict(float)

    for author, value in doa_list:
        dict_unique_author[author] += value  # sum up the doa values for authors that have multiple commits -> merge same authors

    doa_list_unique_author = list(
        map(list, dict_unique_author.items())
    )  # convert the dict to list
    doa_unique_author_values = np.array(doa_list_unique_author)[:, 1].astype(np.float)  # extract the doa values from the list and convert to float

    # if only one author assign doa of 1 else calculate the normalized doa values
    if len(doa_unique_author_values) == 1:
        doa_list_unique_author[0].append(1)
        return doa_list_unique_author  # return doa when only 1 author
    else:
        return normalize_doa_list(
            doa_list_unique_author, doa_unique_author_values
        )  # return normalized doa list


def first_author_file(commit_history):
    """
    Return first author from the commit history
    """

    first_author = commit_history[-1][2]

    return first_author


def normalize_doa_list(doa_list, doa_values):
    """
    Calculate the degree of authorship for a source file for every author from a commit history

    Parameters
    ----------
    doa_list : list of lists - [[author, doa]]
        List of author and doa values.
    doa_values : list - [doa]
        List of doa values.

    Returns
    -------
    normalized_list : list of lists - [[author, doa, normalized doa]]
        Normalized degree-of-authorship list.
    """

    normalized_values = (doa_values - np.min(doa_values)) / (
        np.max(doa_values) - np.min(doa_values)
    )
    normalized_list = []

    for author, degree_of_authorship in zip(doa_list, normalized_values):
        normalized_list.append([author[0], author[1], degree_of_authorship])

    return normalized_list


def get_authors_file(list_doa, filename):
    """
    Return the authors of a file.
    """

    # set threshold parameters
    th_norm_doa = 0.75
    th_abs_doa = 20

    authors_file = [filename]

    # loop through the doa list and extract the authors of a file if the doa values exceed a certain threshold or the author has first authorship
    for item in list_doa:
        if item[1] >= th_abs_doa and item[2] > th_norm_doa or item[2] == 1:
            authors_file.append(item[0])  # append the author to the list

    return authors_file


def calculate_tf(local_path_repo, cur_wd, dict_authors_file):
    """
    Calculate the truck factor for a repository and writes the truckfactor statistics to a csv file.
    Loop through the list of authors
        - get the coverage -> percentage of covered files by the current authors in the authors_list
        - remove the top author (author with the most covered files)
        - increase the truck factor by 1
        - if the coverage is less than 0.5 stop the iteration and return the truck factor and tf statistics

    Returns
    -------
    truckfactor - int
        Number of truckfactor developers.
    """

    authors_list = list(dict_authors_file.keys())  # create a list with the authors
    author_num_files = dict(
        map(lambda author: (author, len(dict_authors_file[author])), authors_list)
    )  # create a dictionary with the authors and the number of files they authored
    num_changed_files = get_num_unique_files(
        dict_authors_file
    )  # get the unique total number of files authored
    tf_stats = [
        ["author", "covered_files", "pct_covered_files", "truck_factor"]
    ]  # create a list which will hold the truckfactor statistics

    repo_name = local_path_repo.rpartition("/")[-1]  # extract the repo name from the repo path
    dict_truckfactor = {"Repository" : repo_name, "Truckfactor_authors" : []}
    truckfactor = 0
    while authors_list:
        coverage = get_coverage(
            authors_list, dict_authors_file, num_changed_files
        )  # get file coverage
        if coverage < 0.5:
            break

        top_author = find_top_author(authors_list, dict_authors_file)  # get top author
        authors_list.remove(top_author)  # remove top author
        top_author_number_files = author_num_files.get(top_author)  # get the number of files from the removed author for the tf statistics
        tf_stats.append(
            [top_author,top_author_number_files,top_author_number_files / num_changed_files,1,]
        )  # add truckfactor statistics for the tf developers
        truckfactor += 1

        #append the truckfactor results to dictionary
        dict_truckfactor.update({"Truckfactor":truckfactor})
        dict_truckfactor["Truckfactor_authors"].append(top_author)

    # add the truckfactors statistics for non tf developers
    for author in authors_list:
        author_number_files = author_num_files.get(author)
        tf_stats.append(
            [author, author_number_files, author_number_files / num_changed_files, 0]
        )

    write_to_csv(local_path_repo, cur_wd, tf_stats, repo_name)  # write the tf statistics to a csv file

    return dict_truckfactor

'''
if (file_authors[i] in dict_files_authors):  # if author is in dictionary append the filenames
    dict_files_authors[file_authors[i]].append(file_authors[0])
else:
    dict_files_authors[file_authors[i]] = [file_authors[0]] 
'''

def get_num_unique_files(dict_authors_file):
    """
    Return the unique total number of files authored from dict_authors_file.
    """

    changed_files_unchained = list(dict_authors_file.values())
    changed_files_chained = list(itertools.chain.from_iterable(changed_files_unchained))
    num_changed_files = len(set(changed_files_chained))

    return num_changed_files


def get_coverage(authors_list, dict_authors_file, num_changed_files):
    """
    Return the pct of covered files by a set of authors.
    """

    covered_files = set()

    for author in authors_list:
        covered_files.update(dict_authors_file[author])

    return len(covered_files) / num_changed_files


def find_top_author(authors_list, dict_authors_file):
    """
    Return the top author from a set of authors.

    Parameters
    ----------
    authors_list : list - [author]
        List of authors.
    dict_authors_file : dict - {author: [filename]}
        Dictionary that stores the authors and their authored files.
    """

    author_num_files = dict(
        map(lambda author: (author, len(dict_authors_file[author])), authors_list)
    )
    top_author = max(author_num_files, key=author_num_files.get)

    return top_author


def write_to_csv(local_path_repo, cur_wd, tf_stats, repo_name):
    """
    Write the truckfactor statistics to a csv file.
    """

    pathlib_curwd = Path(cur_wd)

    directory_path = pathlib_curwd.joinpath("truckfactor_results")

    try:
        os.makedirs(directory_path)  # create directory if not already existent
    except:
        pass

    csv_file = directory_path.joinpath("truckfactor_stats_" + repo_name + ".csv")
    #csv_file = (
    #    directory_path + "\\truckfactor_stats_" + repo_name + ".csv"
    #)  # path to csv file

    # write the truck factor stats to csv file
    with open(csv_file, "w", encoding="utf-8") as f:
        writer = csv.writer(f)
        writer.writerows(tf_stats)



import os
import subprocess
import networkx as nx
from pathlib import Path

def save_network_adjlist(local_path_repo, cur_wd, start_date, end_date, year, list_files):
    repo_name = local_path_repo.rpartition("/")[-1]
    destination = Path("./data/developer_networks/" + year + "/" + repo_name + "_adjacency_list.adjlist")

    if not destination.exists():
        dict_file_dev = create_dict_file_dev(local_path_repo, cur_wd, start_date, end_date, list_files)
        set_contributors = get_set_contributors(dict_file_dev)
        adjacency_list = create_adjacency_list(set_contributors, dict_file_dev)

        graph_repo = nx.Graph(adjacency_list)

        nx.write_adjlist(graph_repo, destination)


def create_dict_file_dev(local_path_repo, cur_wd, start_date, end_date, list_files):
    dict_file_dev = {}

    for source_file in list_files:
        list_contributors = get_contributors_file(local_path_repo, cur_wd, source_file, start_date, end_date)
        if len(list_contributors) <= 1:
            continue
        dict_file_dev[source_file] = list_contributors

    return dict_file_dev


def get_contributors_file(local_path_repo, cur_wd, filename, start_date, end_date):

    git_cmd = [
        "git", "log", "--after={" + start_date + "}", 
        "--before={" + end_date + "}", "--format=%an",
        filename
        ]

    pathlib_curwd = Path(cur_wd)
    pathlib_repo = pathlib_curwd.joinpath(local_path_repo)

    os.chdir(str(pathlib_repo))
    process = subprocess.run(git_cmd, capture_output=True, encoding="utf-8", check=True)
    os.chdir(str(pathlib_curwd))  # change back to the current working directory

    log_shortstat = process.stdout.split("\n")

    list_contributors_file = list(set(filter(None, log_shortstat)))
    
    return list_contributors_file


def get_set_contributors(dict_file_dev):
    set_contributors = set()

    for list_cont in dict_file_dev.values():
        set_contributors.update(list_cont)
    
    return set_contributors


def create_adjacency_list(set_contributors, dict_file_dev):
    adjacency_list = {}

    for contributor in set_contributors:
        for nodes in dict_file_dev.values():
            if contributor in nodes:
                connected_nodes = [node for node in nodes if node != contributor]
                if (contributor in adjacency_list):
                    adjacency_list[contributor].update(connected_nodes)
                else:
                    adjacency_list[contributor] = set(connected_nodes)

    return adjacency_list


def calculate_closeness_centrality(G, number):
    closeness_all = nx.closeness_centrality(G)
    closeness = sorted(closeness_all.items(),key=lambda x: x[1],reverse=True)[0:number]
    return closeness



from setuptools import setup
import os
import sys

_here = os.path.abspath(os.path.dirname(__file__))

if sys.version_info[0] < 3:
    with open(os.path.join(_here, 'README.rst')) as f:
        long_description = f.read()
else:
    with open(os.path.join(_here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()

version = {}
with open(os.path.join(_here, 'git_truckfactor', 'version.py')) as f:
    exec(f.read(), version)

setup(
    name='git-truckfactor',
    version=version['__version__'],
    description=('Calculate code contribution metrics for git repositories'),
    long_description=long_description,
    author='Philipp Rettig',
    author_email='ph.rettig13@gmail.com',
    url='https://github.com/PhilRett/git-truckfactor',
    license='',
    packages=['git_truckfactor'],
    install_requires=['numpy', 'networkx', 'plotly'],
    include_package_data=True,
    classifiers=[   
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.9'],
    )


"""
Module for cloning and extracting metrics from a git repository.
"""
import subprocess
import os
import shlex
import re
import pandas as pd
import plotly.graph_objects as go
from datetime import datetime
from pathlib import Path


def clone_repository(url_repo, path_to):
    """
    Clone a git repository to a specified path.

    Parameters
    ----------
    url_repo : str
        URL address of git repository.
    path_to : str
        Local path to destination directory.
    """

    #specify the git command, split the string and put into list
    git_cmd = shlex.split('git clone ' + url_repo + " " + path_to)

    destination = Path(path_to)

    if not destination.exists():
        count = 1
        while count <= 3:
            try:
                process = subprocess.run(git_cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, check = True)
                process #run the git command
                break
            except:
                count += 1
                continue

    return path_to


def get_info(local_path_repo, cur_wd):
    """
    Return commit hash, author, author date in 'YYYY-MM-DD' format, author email address, commit comment, insertions and deletions.

    Parameters
    ----------
    local_path_repo : str
        Local path to the directory of the git repository.
    cur_wd : str
        Current working directory.

    Returns
    -------
    commit_info : list of lists - [['Commit', hash, author, date, email, commit comment, [insertions, deletions, total changes]]]
    """

    git_cmd = [
        "git", "log", "--pretty=format:Commit%n%H%n%an%n%as%n%ae%n%s%n", 
        "--shortstat"
        ]

    pathlib_curwd = Path(cur_wd)
    pathlib_repo = pathlib_curwd.joinpath(local_path_repo)

    os.chdir(str(pathlib_repo))
    process = subprocess.run(git_cmd, capture_output=True, encoding="utf-8", check=True)
    os.chdir(str(pathlib_curwd))  # change back to the current working directory

    log_shortstat = process.stdout.split("\n")
# append the ouptut of git log to a list, each commit stat can be separated by the entry "Commit"
    list_commit_info = []
    i_sep_cur = 0
    i_sep_prev = 0
    for i in range(1, len(log_shortstat)):  # loop through list
        if (log_shortstat[i] == "Commit"):  # if 'Commit' is found set the index for the previous and current separators
            i_sep_prev = i_sep_cur
            i_sep_cur = i
            list_commit_info.append(
                log_shortstat[i_sep_prev : i_sep_cur - 1]
            )  # append all items between string "commit"
    list_commit_info.append(
        log_shortstat[i_sep_cur:-1]
    )  # append items after last string "commit"

    list_commit_info = [
        list(filter(None, commit)) for commit in list_commit_info
    ]  # filter None values

    # extract the insertions and deletions
    for commit in list_commit_info:
        try:
            re_insert = re.search(r"(\d+|$)(?=\s*insert)", commit[6])
            insertions = re_insert.group(0) if re_insert else 0
            re_delet = re.search(r"(\d+|$)(?=\s*delet)", commit[6])
            deletions = re_delet.group(0) if re_delet else 0
            commit[6] = [insertions, deletions, int(insertions) + int(deletions)]
        except:
            # if no changes are made, the insertions and deletions are not displayed, therefore we add them if no changes can be detected
            commit.append([0, 0, 0])

    return list_commit_info


def commits_per_user(list_commit_info):
    """
    Return a dictionary that holds the people that commited on the specified repositories and the number of commits

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()

    Returns
    -------
    dict_commits_user : dict - {'author': number of commits}
    """
    
    dict_commits_user = {}

    for commit in list_commit_info:
        author = commit[2]
        if (author in dict_commits_user):
            dict_commits_user[author] += 1
        else:
            dict_commits_user[author] = 1

    return dict_commits_user


def number_commits(list_commit_info):
    """
    Return a dictionary that holds the total number of commits and contributors of a repository

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()

    Returns
    -------
    dict_commits_stats : dict - {'Total number of commits': num_commits
                                 'Number of contributors': num_contributors}
    """
    dict_commits_user = commits_per_user(list_commit_info)
    num_commits = sum(dict_commits_user.values()) #sum the values in the dictionary
    num_contributors = len(dict_commits_user) #length of the dictionary -> number of people
    
    #create a dictionary to store the values
    dict_commits_stats = {}
    dict_commits_stats["Total number of commits:"] = num_commits
    dict_commits_stats["Number of contributors"] = num_contributors

    return dict_commits_stats


def list_contributors(list_commit_info):
    """
    Return a list of contributors

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()

    Returns
    -------
    list_contributors : list - [contributors]
    """
    list_contributors = list(set([commit[2] for commit in list_commit_info]))

    return list_contributors


def list_first_contributions(list_commit_info):
    """
    Return a list that holds the date of the initial contributions

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()

    Returns
    -------
    list_first_contribution : list
    """
    dict_first_contribution = {}

    for commit in list_commit_info:
        contributor = commit[2]
        date = datetime.strptime(commit[3], "%Y-%m-%d")

        if (contributor in dict_first_contribution):
            if date < dict_first_contribution[contributor]:
                dict_first_contribution[contributor] = date
        else:
            dict_first_contribution[contributor] = date
    
    list_first_contribution = list(map(list, dict_first_contribution.items()))

    return list_first_contribution


def plot_new_contributors_month(list_commit_info):
    """
    Create a plot that shows new contributors per month

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()
    """
    
    list_first_contribution = list_first_contributions(list_commit_info)
    
    #transform commit_info to a pandas dataframe
    df_first_contribution = pd.DataFrame(list_first_contribution, columns = ['contributor', 'date_first_contribution'])

    #group commit_info by month
    first_cont_month = df_first_contribution.groupby([df_first_contribution['date_first_contribution'].dt.strftime('%m-%Y')]).agg({'contributor': pd.Series.nunique}) #group by month and year, count unique values using aggregate function
    first_cont_month.index = pd.to_datetime(first_cont_month.index) #convert index to datetime type
    first_cont_month = first_cont_month.sort_index() #sort by index/date

    #Create a plot that shows the unique contributors per month
    month_year = first_cont_month.index

    count_cont_month = first_cont_month['contributor']

    fig = go.Figure()

    fig.add_trace(go.Scatter(x=month_year, y=count_cont_month, name='Count of new contributors per month',
                         line=dict(color='royalblue', width=2)))

    fig.update_layout(title='Count of new contributors per month over time',
                   xaxis_title='Date',
                   yaxis_title='Count of  newcontributors')


    fig.show()


def lines_of_code_added(list_commit_info):
    """
    Return a dictionary that holds the contributors and their delivered lines of code

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()

    Returns
    -------
    dict_loc_user : dict - {"author" : loc_added}
    """

    dict_loc_user = {}

    for commit in list_commit_info:
        author = commit[2]
        loc_added = int(commit[-1][0])
        if (author in dict_loc_user):
            dict_loc_user[author] += loc_added
        else:
            dict_loc_user[author] = loc_added

    return dict_loc_user


def plot_unique_contributors_month(list_commit_info):
    """
    Create a plot that shows the unique contributors per month

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()
    """

    commit_info = [commit[1:5] for commit in list_commit_info]

    #transform commit_info to a pandas dataframe
    df_commit_info = pd.DataFrame(commit_info, columns = ['commit hash', 'author', 'date', 'email'])
    df_commit_info['date']= pd.to_datetime(df_commit_info['date'])

    #group commit_info by month
    cont_month = df_commit_info.groupby([df_commit_info['date'].dt.strftime('%m-%Y')]).agg({'author': pd.Series.nunique}) #group by month and year, count unique values using aggregate function
    cont_month.index = pd.to_datetime(cont_month.index) #convert index to datetime type
    cont_month = cont_month.sort_index() #sort by index/date

    #Create a plot that shows the unique contributors per month
    month_year = cont_month.index
    count_cont_month = cont_month['author']

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=month_year, y=count_cont_month, name='Count of contributors per month',
                         line=dict(color='royalblue', width=2)))

    fig.update_layout(title='Count of unique contributors per month over time',
                      xaxis_title='Date',
                      yaxis_title='Count of contributors')
    fig.show()


def plot_unique_contributors_quarter(list_commit_info):
    """
    Create a plot that shows the unique contributors per quarter

    Parameters
    ----------
    list_commit_info : list
        list that holds information about commits, which is the list retrieved from function get_info()
    """

    commit_info = [commit[1:5] for commit in list_commit_info]

    #transform commit_info to a pandas dataframe
    df_commit_info = pd.DataFrame(commit_info, columns = ['commit hash', 'author', 'date', 'email'])
    df_commit_info['date']= pd.to_datetime(df_commit_info['date'])

    #group df_commit_info by quarter
    cont_quarter = df_commit_info.groupby(df_commit_info['date'].dt.to_period('Q')).agg({'author': pd.Series.nunique}) #group by month and year count 
    cont_quarter = cont_quarter.sort_index().to_timestamp() #convert the period type to a timestamp to be able to plot it with plotly

    #Create a plot that shows the unique contributors per quarter
    quarter = cont_quarter.index
    count_cont_quarter = cont_quarter['author']

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=quarter, y=count_cont_quarter, name='Count of contributors per quarter',
                         line=dict(color='firebrick', width=2)))

    fig.update_layout(title='Count of unique contributors per quarter over time',
                   xaxis_title='Date',
                   yaxis_title='Count of contributors')
    fig.show()